IncludeFile "SPTI.pb"

Procedure.l ExecCMD(DrvID.s, Array cmd.b(1), CmdLen.w, WFE.l, eDir.l, BufPtr.l, buflen.l, Timeout.l = 5)
	DataDir.l
	If eDir = #SRB_DIR_IN
		DataDir = #SCSI_IOCTL_DATA_IN
	Else
		DataDir = #SCSI_IOCTL_DATA_OUT
	EndIf
	ProcedureReturn SPTICmd(DrvID, cmd(), CmdLen, WFE, DataDir, BufPtr, buflen, Timeout)
EndProcedure

;Read a Mode Page
Procedure.l CDRomModeSense10(DrvID.s, MP.b, PtrBuffer.l, BufferLen.l)
	Dim cmd.b(9)
	cmd(0) = $5A                     ;  ' MODE SENSE 10 OpCode
	cmd(2) = MP & $3F                ;' Mode Page
	cmd(7) = BufferLen >> 8   ; ' allocation length
	cmd(8) = BufferLen & $FF   ;      ' allocation length
	ProcedureReturn ExecCMD(DrvID, cmd(), 10, #False, #SRB_DIR_IN, PtrBuffer, BufferLen + 1)
EndProcedure

;send a Mode Page
Procedure.l CDRomModeSelect10(DrvID.s, PtrBuffer.l, BufferLen.l)
	Dim cmd.b(9)
	cmd(0) = $55                     ; MODE SELECT10 OpCode
	cmd(1) = $10                     ; PF = 1 (Page Format)
	cmd(7) = (BufferLen >> 8) & $FF  ; allocation length
	cmd(8) = BufferLen & $FF         ; allocation length
	ProcedureReturn ExecCMD(DrvID, cmd(), 10, #True, #SRB_DIR_OUT, PtrBuffer, BufferLen)
EndProcedure

;send new write parameters page
Procedure.l CDRomWriteParams(strDrv.s, TestMode.l, BURNProof.l, TrackPause.l, WriteType.b, TrackMode.b, DataBlockType.b, MultiSession.l)
	Dim bufData.b(60)
	PS.b
	i.w

	bufData(1) = 58     ; length
	bufData(8) = 5      ; Page Code
	bufData(9) = $32     ; Page length
	
	;WriteType, Test-Mode and Burn Proof
	bufData(10) = WriteType | (TestMode * $10) | (BURNProof * $40)

	;Track-Mode, Multi-Session
	bufData(11) = TrackMode | (MultiSession * $C0)
	
	;Data-Mode (Mode 1: 2048 Bytes User-Data)
	bufData(12) = DataBlockType
	
	;Track Pause: 150 sectors (frames) = 2 seconds
	bufData(22) = (TrackPause >> 8) & $FF
	bufData(23) = TrackPause & $FF
	
	;send new WPP
	ProcedureReturn CDRomModeSelect10(strDrv, @bufData(), 60)
EndProcedure

;Read Drive Features
Procedure.l ReadDriveAttributes(DrvID.s, Exec_Command.l, Exec_len.l, PtrBuffer.l, BufferLen.l)
	Dim cmd.b(11)
	Exec_Command2.l
	cmd(0) = $46
	cmd(1) = 2
	cmd(9) = 0
	If Exec_Command > 255
	Exec_Command2 = 1
	Exec_Command & $FF
	EndIf
	cmd(2) = Exec_Command2
	cmd(3) = Exec_Command
	cmd(4) = 0
	cmd(5) = 0
	cmd(6) = 0
	cmd(7) = 0
	cmd(8) = Exec_len
	ProcedureReturn ExecCMD(DrvID, cmd(), 12, #True, #SRB_DIR_IN, PtrBuffer, BufferLen, 10)
EndProcedure

Procedure.l From4Byte(Array header.b(1))
	ProcedureReturn (header(0) << 24) | (header(1) << 16) | (header(2) << 8) | header(3)
EndProcedure

;Read drive features
Procedure.l CDRomGetConfiguration(strDrv.s, StartFeature.b, RT.b, PtrBuffer.l, buflen.l)
    Dim cmd.b(9)
    cmd(0) = $46                   			; GET CONFIGURATION Op-Code
    cmd(1) = RT                     		; RT Byte
    cmd(2) = (StartFeature >> 8) & $FF    ; start feature
    cmd(3) = StartFeature & $FF  			; start feature
    cmd(7) = (buflen >> 8) & $FF   			; allocation length
    cmd(8) = buflen & $FF        			; allocation length
    ProcedureReturn ExecCMD(strDrv, cmd(), 10, #False, #SRB_DIR_IN, PtrBuffer, buflen + 1, 10)
EndProcedure

Procedure.s CDRomGetInterface(strDrv.s)
	Dim buffer.b(15): Dim cmd.b(9)
	inquiry.t_InqDat
	;try to read the drive's core feature
	CDRomGetConfiguration(strDrv, 1, 2, @buffer(), ArraySize(buffer()))
	;determine the interface from it
	tmp.l = (buffer(12) << 24) | (buffer(13) << 16) | (buffer(14) << 8) | buffer(15)
	Select tmp
		Case 0: ProcedureReturn "VIRTUAL"
		Case 1: ProcedureReturn "SCSI"
		Case 2: ProcedureReturn "ATAPI"
		Case 3, 4: ProcedureReturn "IEEE"
		Case 7: ProcedureReturn "SATA"
	EndSelect
	;if it didn't work, try INQUIRY
	cmd(0) = $12               ; Inquiry OpCode
	cmd(4) = SizeOf(inquiry) - 1  ; allocation length
	If ExecCMD(strDrv, cmd(), 10, #False, #SRB_DIR_IN, @inquiry, SizeOf(inquiry), 10)
		If inquiry\ver = 0: ProcedureReturn "ATAPI": EndIf
		If inquiry\ver >= 2: ProcedureReturn "VIRTUAL": EndIf
	EndIf
	ProcedureReturn "UNKNOWN"
EndProcedure

Procedure GetOpticalDriveInfo(strDrvID.s, *DriveInfo.OpticalDriveInfo)
	aa.t_HAID
	SPTIGetHAID(strDrvID,@aa)
	*DriveInfo\BusNo = aa\HA
	*DriveInfo\TargetNo = aa\ID
	*DriveInfo\LunNo = aa\DEV_Lun
	inq.t_InqDat     ; Inquiry structure
	Dim cmd.b(5)         ; CDB
	cmd(0) = $12               ; 0x12 Inquiry (6 Byte)
	cmd(4) = SizeOf(inq) - 1       ; allocation length
	;execute the command (inquiry needs to be supported by every CD/DVD-ROM)
	ExecCMD(strDrvID, cmd(), 6, #True, #SRB_DIR_IN, @inq, SizeOf(inq))
	
	*DriveInfo\VendorID = PeekS(@inq\VID,8);StrConv(inq.VID, vbUnicode)
	*DriveInfo\ProductID = PeekS(@inq\PID,16)
	*DriveInfo\Firmware = PeekS(@inq\PVER,4)
	*DriveInfo\ExData = PeekS(@inq\FWVER,21)
	;take features
	mmc.t_MMC
	CDRomModeSense10(strDrvID, $2A, @mmc, SizeOf(mmc) - 1)

	With *DriveInfo\ReadFeatures
		\CDR = mmc\ReadSupported & 1
		\CDRW = (mmc\ReadSupported >> 1) & 1
		\DVDROM = (mmc\ReadSupported >> 3) & 1
		\DVDR = (mmc\ReadSupported >> 4) & 1
		\DVDRW = (mmc\ReadSupported >> 4) & 1; I didn't find a way to determine it, but I do not think that there is a device that can read DVD-R and not DVD-RW. So, if DVD-R can be read, DVD-RW can be read too.
		\DVDRAM = (mmc\ReadSupported >> 5) & 1
		\CDDARawRead = (mmc\misc[1] >> 0) & 1
		\Mode2Form1 = (mmc\misc[0] >> 4) & 1
		\Mode2Form2 = (mmc\misc[0] >> 5) & 1
		\MultiSession = (mmc\misc[0] >> 6) & 1
		\ISRC = (mmc\misc[1] >> 5) & 1
		\UPC = (mmc\misc[1] >> 6) & 1
		\BC = (mmc\misc[1] >> 7) & 1
		\SubChannels = (mmc\misc[1] >> 2) & 1
		\SubChannelsCorrected = (mmc\misc[1] >> 3) & 1
		\SubChannelsFromLeadIn = (mmc\misc[3] >> 5) & 1
		\C2ErrorPointers = (mmc\misc[1] >> 4) & 1
	EndWith
	;some write features
	With *DriveInfo\WriteFeatures
		\CDR = (mmc\WriteSupported >> 0) & 1
		\CDRW = (mmc\WriteSupported >> 1) & 1
		\TestMode = (mmc\WriteSupported >> 2) & 1
		\DVDR = (mmc\WriteSupported >> 4) & 1
		\DVDRAM = (mmc\WriteSupported >> 5) & 1
		\BURNProof = (mmc\misc[0] >> 7) & 1
	EndWith
	
	With *DriveInfo\Speeds
		\MaxRSpeed = (mmc\MaxReadCDSpeed[0] << 8) | mmc\MaxReadCDSpeed[1]
		\CurrRSpeed = (mmc\CurrReadCDSpeed[0] << 8) | mmc\CurrReadCDSpeed[1]
		\MaxWCDRSpeed = (mmc\MaxWriteCDRSpeed[0] << 8) | mmc\MaxWriteCDRSpeed[1]
		\CurrWSpeed = (mmc\CurrWriteSpeed[0] << 8) | mmc\CurrWriteSpeed[1]
		If \CurrRSpeed > \MaxRSpeed: \CurrRSpeed = \MaxRSpeed:EndIf
		If \CurrWSpeed > \MaxWCDRSpeed: \CurrWSpeed = \MaxWCDRSpeed: EndIf
	EndWith
	
	;the rest of reading and writing features need extra command executions
	With *DriveInfo\WriteFeatures
		Dim header.b(127): ar.w
		; WRITE: DVD-R, DVD-RW, DVD-R DL
		ReadDriveAttributes(strDrvID, $2F, 16, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 12
			\DVDR = #True
			If (header(12) & 2) = 2: \DVDRW = #True:EndIf
			If (header(12) & 4) = 4: \DVD_R_DL = #True: EndIf
		EndIf
		;WRITE: DVD-R DL
		ReadDriveAttributes(strDrvID, $33, 16, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 12
			\LAYER_JUMP = #True
			\DVD_R_DL = #True
		EndIf
		;WRITE: DVD+R, READ: DVD+R
		ReadDriveAttributes(strDrvID, $2B, 16, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 12
			*DriveInfo\ReadFeatures\DVD_P_R = #True
			If (header(12) & 1) = 1: \DVD_P_R = #True: EndIf
		EndIf
		;WRITE: DVD+RW, READ: DVD+RW
		ReadDriveAttributes(strDrvID, $2A, 16, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 12
			*DriveInfo\ReadFeatures\DVD_P_RW = #True
			If (header(12) & 1) = 1: \DVD_P_RW = #True: EndIf
		EndIf
		;WRITE:DVD+R DL, READ: DVD+R DL
		ReadDriveAttributes(strDrvID, $3B, 16, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 12
			*DriveInfo\ReadFeatures\DVD_P_R_DL = #True
			If (header(12) & 1) = 1: \DVD_P_R_DL = #True: EndIf
		EndIf
		;WRITE: DVD+RW DL, READ: DVD+RW DL
		ReadDriveAttributes(strDrvID, $3A, 16, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 12
			*DriveInfo\ReadFeatures\DVD_P_RW_DL = #True
			If (header(12) & 1) = 1: \DVD_P_RW_DL = #True: EndIf
		EndIf
		;Blue Ray WRITE
		ReadDriveAttributes(strDrvID, $41, 32, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 28
			If (header(16) Or header(17) Or header(18) Or header(19) Or header(20) Or header(21) Or header(22) Or header(23))
				\BD_R = #True
			EndIf
			If (header(24) Or header(25) Or header(26) Or header(27) Or header(28) Or header(29) Or header(30) Or header(31))
				\BD_RW = #True
			EndIf
		EndIf
		;Blue Ray READ
		ReadDriveAttributes(strDrvID, $40, 40, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 36
			If (header(16) Or header(17) Or header(18) Or header(19) Or header(20) Or header(21) Or header(22) Or header(23))
			 *DriveInfo\ReadFeatures\BD_RW = #True
			EndIf
			If (header(24) Or header(25) Or header(26) Or header(27) Or header(28) Or header(29) Or header(30) Or header(31))
			 *DriveInfo\ReadFeatures\BD_R = #True
			EndIf
			If (header(32) Or header(33) Or header(34) Or header(35) Or header(36) Or header(37) Or header(38) Or header(39))
			 *DriveInfo\ReadFeatures\BD_ROM = #True
			EndIf
		EndIf
		;HD DVD READ
		ReadDriveAttributes(strDrvID, $50, 16, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 12
			*DriveInfo\ReadFeatures\HD_DVD_ROM = #True
			If (header(12) & 1): *DriveInfo\ReadFeatures\HD_DVD_R = #True: EndIf
			If (header(14) & 1): *DriveInfo\ReadFeatures\HD_DVD_RAM = #True: EndIf
		EndIf
		;HD DVD WRITE
		ReadDriveAttributes(strDrvID, $51, 16, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 12
			If (header(12) & 1): \HD_DVD_R = #True: EndIf
			If (header(14) & 1): \HD_DVD_RAM = #True: EndIf
		EndIf
		;other features
		;CD Track At Once
		ReadDriveAttributes(strDrvID, $2D, 16, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 8
			\CD_TAO = #True
		EndIf
		;Defect Management
		ReadDriveAttributes(strDrvID, $24, 16, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 8
			\Def_Management = #True
		EndIf
		;Multi Read
		ReadDriveAttributes(strDrvID, $1D, 16, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 8
			*DriveInfo\ReadFeatures\MultiRead = #True
		EndIf
		;CD-RW CAV write
		ReadDriveAttributes(strDrvID, $27, 16, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 8
			\CDRWCAV_Write = #True
		EndIf
		;Mount Rainier Write
		ReadDriveAttributes(strDrvID, $28, 16, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 8
			\MRW = #True
		EndIf
		;DDCD read
		ReadDriveAttributes(strDrvID, $30, 16, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 8
			*DriveInfo\ReadFeatures\DDCD_Read = #True
		EndIf
		;DDCD-R write
		ReadDriveAttributes(strDrvID, $31, 16, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 8
			\DDCDR_Write = #True
		EndIf
		;DDCD-RW write
		ReadDriveAttributes(strDrvID, $32, 16, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 8
			\DDCDRW_Write = #True
		EndIf
		;power management
		ReadDriveAttributes(strDrvID, $100, 12, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 8 
			*DriveInfo\OtherFeatures\PowerManagement = #True
		EndIf
		;embedded changer
		ReadDriveAttributes(strDrvID, $102, 12, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 8 
			*DriveInfo\OtherFeatures\EmbeddedChanger = #True
		EndIf
		;CD Audio Analog Play
		ReadDriveAttributes(strDrvID, $103, 12, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 8 
			*DriveInfo\OtherFeatures\CDAudioAnalogPlay = #True
		EndIf
		;Microcode Upgrade
		ReadDriveAttributes(strDrvID, $104, 12, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 8
			*DriveInfo\OtherFeatures\MicrocodeUpgrade = #True
		EndIf		
		;TimeOut
		ReadDriveAttributes(strDrvID, $105, 12, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 8 
			*DriveInfo\OtherFeatures\TimeOut = #True
		EndIf
		;DVD CSS
		ReadDriveAttributes(strDrvID, $106, 12, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 8 
			*DriveInfo\OtherFeatures\DVD_CSS = #True
		EndIf
		;Real Time Streaming
		ReadDriveAttributes(strDrvID, $107, 12, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 8 
			*DriveInfo\OtherFeatures\RealTimeStreaming = #True
		EndIf
		;Logical Unit Serial Number
		ReadDriveAttributes(strDrvID, $108, 12, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 8
			*DriveInfo\OtherFeatures\LogicalUnitSerialNumber = #True
		EndIf
		;Disc Control Blocks
		ReadDriveAttributes(strDrvID, $10A, 12, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 8 
			*DriveInfo\OtherFeatures\DiscControlBlocks = #True
		EndIf
		;DVD CPRM
		ReadDriveAttributes(strDrvID, $10B, 12, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 8 
			*DriveInfo\OtherFeatures\DVD_CPRM = #True
		EndIf
		;Firmware Date
		ReadDriveAttributes(strDrvID, $10C, 12, @header(), ArraySize(header())+1)
		If From4Byte(header()) >= 8 
			*DriveInfo\OtherFeatures\FirmwareDate = #True
		EndIf
	EndWith
	With *DriveInfo\OtherFeatures
		\LockMedia = mmc\misc[2] & 1
		\AnalogAudio = mmc\misc[0] & 1
		\JitterCorrection = (mmc\misc[1] >> 1) & 1
		\BufferSize = (mmc\BufferSize[0] << 8) | mmc\BufferSize[1]
		\VolumeLevels = (mmc\NumVolLevels[0] << 8) | mmc\NumVolLevels[1]
		Select (mmc\misc[2] >> 5) & 7 
			Case 0
				\LoadingMechanism = "Caddy"
			Case 1
				\LoadingMechanism = "Tray"
			Case 2
				\LoadingMechanism = "Popup"
			Case 4, 5
				\LoadingMechanism = "Changer"
			Default
				\LoadingMechanism = "Unknown"
		EndSelect
   EndWith
   ;writing modes
   With *DriveInfo\WriteFeatures
		If CDRomWriteParams(strDrvID, #False, #False, 150, 1, 4, 8, #False)
			\WriteModes\TAO = #True
		EndIf
		If CDRomWriteParams(strDrvID, #True, #False, 150, 1, 4, 8, #False)
			\WriteModes\TAOTest = #True
		EndIf
		If CDRomWriteParams(strDrvID, #False, #False, 150, 2, 0, 0, #False)
			\WriteModes\SAO = #True
		EndIf
		If CDRomWriteParams(strDrvID, #True, #False, 150, 2, 0, 0, #False)
			\WriteModes\SAOTest = #True
		EndIf
		If CDRomWriteParams(strDrvID, #False, #False, 150, 2, 0, 2, #False)
			\WriteModes\SAOR96P = #True
		EndIf
		If CDRomWriteParams(strDrvID, #True, #False, 150, 2, 0, 2, #False)
			\WriteModes\SAOR96PTest = #True
		EndIf
		If CDRomWriteParams(strDrvID, #False, #False, 150, 2, 0, 3, #False)
			\WriteModes\SAOR96R = #True
		EndIf
		If CDRomWriteParams(strDrvID, #True, #False, 150, 2, 0, 3, #False)
			\WriteModes\SAOR96RTest = #True
		EndIf
		If CDRomWriteParams(strDrvID, #False, #False, 150, 3, 0, 1, #False)
			\WriteModes\Raw16 = #True
		EndIf
		If CDRomWriteParams(strDrvID, #True, #False, 150, 3, 0, 1, #False)
			\WriteModes\Raw16Test = #True
		EndIf
		If CDRomWriteParams(strDrvID, #False, #False, 150, 3, 0, 2, #False)
			\WriteModes\Raw96P = #True
		EndIf
		If CDRomWriteParams(strDrvID, #True, #False, 150, 3, 0, 2, #False)
			\WriteModes\Raw96PTest = #True
		EndIf
		If CDRomWriteParams(strDrvID, #False, #False, 150, 3, 0, 3, #False)
			\WriteModes\Raw96R = #True
		EndIf
		If CDRomWriteParams(strDrvID, #True, #False, 150, 3, 0, 3, #False)
			\WriteModes\Raw96RTest = #True
		EndIf
	EndWith
	;get drive interface
	*DriveInfo\OtherFeatures\DriveInterface = CDRomGetInterface(strDrvID)
EndProcedure
; IDE Options = PureBasic 4.40 (Windows - x86)
; CursorPosition = 109
; FirstLine = 81
; Folding = --
; EnableXP